import { CommonModule } from '@angular/common';
import { NG_VALUE_ACCESSOR, FormsModule } from '@angular/forms';
import { takeUntil } from 'rxjs/operators';
import { Subject } from 'rxjs';
import { Component, ElementRef, EventEmitter, HostBinding, ViewChild, ViewEncapsulation, ChangeDetectorRef, ComponentFactoryResolver, HostListener, Input, NgZone, Output, ViewContainerRef, Directive, forwardRef, NgModule } from '@angular/core';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/** @enum {number} */
const Key = {
    Backspace: 8,
    Tab: 9,
    Enter: 13,
    Shift: 16,
    Escape: 27,
    Space: 32,
    End: 35,
    Home: 36,
    ArrowLeft: 37,
    ArrowUp: 38,
    ArrowRight: 39,
    ArrowDown: 40,
};
Key[Key.Backspace] = 'Backspace';
Key[Key.Tab] = 'Tab';
Key[Key.Enter] = 'Enter';
Key[Key.Shift] = 'Shift';
Key[Key.Escape] = 'Escape';
Key[Key.Space] = 'Space';
Key[Key.End] = 'End';
Key[Key.Home] = 'Home';
Key[Key.ArrowLeft] = 'ArrowLeft';
Key[Key.ArrowUp] = 'ArrowUp';
Key[Key.ArrowRight] = 'ArrowRight';
Key[Key.ArrowDown] = 'ArrowDown';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/** @type {?} */
const styleProperties = Object.freeze([
    'direction',
    'boxSizing',
    'width',
    'height', 'overflowX',
    'overflowY',
    'borderTopWidth', 'borderRightWidth', 'borderBottomWidth', 'borderLeftWidth', 'borderStyle',
    'paddingTop', 'paddingRight', 'paddingBottom', 'paddingLeft',
    // https://developer.mozilla.org/en-US/docs/Web/CSS/font
    'fontStyle', 'fontVariant', 'fontWeight', 'fontStretch', 'fontSize', 'fontSizeAdjust', 'lineHeight', 'fontFamily',
    'textAlign', 'textTransform', 'textIndent',
    'textDecoration',
    'letterSpacing', 'wordSpacing',
    'tabSize', 'MozTabSize'
]);
/** @type {?} */
const isBrowser = typeof window !== 'undefined';
/** @type {?} */
const isFirefox = isBrowser && ((/** @type {?} */ (window))).mozInnerScreenX != null;
/**
 * @param {?=} element
 * @return {?}
 */
function isInputOrTextAreaElement(element) {
    return element !== null &&
        (element.nodeName === 'INPUT' || element.nodeName === 'TEXTAREA' || element.nodeName === '#text');
}
/**
 * @param {?} value
 * @param {?} start
 * @param {?} end
 * @param {?} insert
 * @return {?}
 */
function spliceString(value, start, end, insert) {
    return value.substring(0, start) + insert + value.substring(end) + '';
}
/**
 * @param {?} value
 * @param {?} mentionMarkup
 * @param {?} textIterator
 * @param {?} markupIterator
 * @param {?} displayTransform
 * @return {?}
 */
function iterateMentionsMarkup(value, mentionMarkup, textIterator, markupIterator, displayTransform) {
    /** @type {?} */
    let match;
    /** @type {?} */
    let start = 0;
    /** @type {?} */
    let currentPlainTextIndex = 0;
    /** @type {?} */
    let regEx = mentionMarkup.regEx;
    regEx.lastIndex = 0;
    while ((match = regEx.exec(value)) !== null) {
        /** @type {?} */
        let display = displayTransform(...match);
        /** @type {?} */
        let substr = value.substring(start, match.index);
        textIterator(substr, start, currentPlainTextIndex);
        currentPlainTextIndex += substr.length;
        markupIterator(match[0], match.index, currentPlainTextIndex, display);
        currentPlainTextIndex += display.length;
        start = regEx.lastIndex;
    }
    if (start < value.length) {
        textIterator(value.substring(start), start, currentPlainTextIndex);
    }
}
/**
 * @param {?} value
 * @param {?} mentionMarkup
 * @param {?} markupIterator
 * @param {?} displayTransform
 * @return {?}
 */
function iterateOnlyMentionsMarkup(value, mentionMarkup, markupIterator, displayTransform) {
    /** @type {?} */
    let match;
    /** @type {?} */
    let start = 0;
    /** @type {?} */
    let currentPlainTextIndex = 0;
    /** @type {?} */
    let regEx = mentionMarkup.regEx;
    regEx.lastIndex = 0;
    while ((match = regEx.exec(value)) !== null) {
        /** @type {?} */
        let display = displayTransform(...match);
        /** @type {?} */
        let substr = value.substring(start, match.index);
        currentPlainTextIndex += substr.length;
        markupIterator(match[0], match.index, currentPlainTextIndex, display);
        currentPlainTextIndex += display.length;
        start = regEx.lastIndex;
    }
}
/**
 * @param {?} value
 * @param {?} mentionMarkup
 * @param {?} indexInPlainText
 * @param {?} toEndOfMarkup
 * @param {?} displayTransform
 * @return {?}
 */
function mapPlainTextIndex(value, mentionMarkup, indexInPlainText, toEndOfMarkup, displayTransform) {
    if (isNaN(indexInPlainText)) {
        return indexInPlainText;
    }
    /** @type {?} */
    let result;
    /** @type {?} */
    const textIterator = (matchString, index, substringPlainTextIndex) => {
        if (typeof result !== 'undefined') {
            return;
        }
        if (substringPlainTextIndex + matchString.length >= indexInPlainText) {
            result = index + indexInPlainText - substringPlainTextIndex;
        }
    };
    /** @type {?} */
    const markupIterator = (matchString, index, mentionPlainTextIndex, display) => {
        if (typeof result !== 'undefined') {
            return;
        }
        if (mentionPlainTextIndex + display.length > indexInPlainText) {
            result = index + (toEndOfMarkup ? matchString.length : 0);
        }
    };
    iterateMentionsMarkup(value, mentionMarkup, textIterator, markupIterator, displayTransform);
    return typeof result !== 'undefined' ? result : value.length;
}
/**
 * @param {?} element
 * @return {?}
 */
function getCaretPosition(element) {
    if (isInputOrTextAreaElement(element)) {
        /** @type {?} */
        let value = element.value;
        return value.slice(0, element.selectionStart).length;
    }
    /** @type {?} */
    let selection = window.getSelection();
    if (selection.rangeCount > 0) {
        /** @type {?} */
        let range = selection.getRangeAt(0);
        return range.startOffset;
    }
    return 0;
}
/**
 * @param {?} element
 * @param {?} position
 * @return {?}
 */
function getCaretCoordinates(element, position) {
    /** @type {?} */
    let coords = { top: 0, left: 0 };
    if (!isBrowser) {
        return coords;
    }
    /** @type {?} */
    const div = document.createElement('div');
    document.body.appendChild(div);
    /** @type {?} */
    let style = div.style;
    /** @type {?} */
    const computed = window.getComputedStyle ? getComputedStyle(element) : ((/** @type {?} */ (element))).currentStyle;
    style.whiteSpace = 'pre-wrap';
    if (element.nodeName !== 'INPUT') {
        style.wordWrap = 'break-word';
    }
    style.position = 'absolute';
    style.visibility = 'hidden';
    styleProperties.forEach(prop => style[prop] = computed[prop]);
    if (isFirefox) {
        if (element.scrollHeight > parseInt(computed.height, 10)) {
            style.overflowY = 'scroll';
        }
    }
    else {
        style.overflow = 'hidden';
    }
    div.textContent = element.value.substring(0, position);
    if (element.nodeName === 'INPUT') {
        div.textContent = div.textContent.replace(/\s/g, '\u00a0');
    }
    /** @type {?} */
    let span = document.createElement('span');
    span.textContent = element.value.substring(position) || '.';
    div.appendChild(span);
    coords = {
        top: span.offsetTop + parseInt(computed['borderTopWidth'], 10),
        left: span.offsetLeft + parseInt(computed['borderLeftWidth'], 10)
    };
    document.body.removeChild(div);
    return coords;
}
/**
 * @param {?} element
 * @param {?} position
 * @return {?}
 */
function setCaretPosition(element, position) {
    if (isInputOrTextAreaElement(element) && element.selectionStart !== undefined) {
        element.focus();
        element.setSelectionRange(position, position);
    }
    else {
        /** @type {?} */
        let range = document.createRange();
        range.setStart(element, position);
        range.collapse(true);
        /** @type {?} */
        let selection = window.getSelection();
        selection.removeAllRanges();
        selection.addRange(range);
    }
}
/**
 * @param {?} str
 * @return {?}
 */
function escapeRegExp(str) {
    return str.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&');
}
/**
 * @param {?} markup
 * @return {?}
 */
function markupToRegExp(markup) {
    /** @type {?} */
    let markupPattern = escapeRegExp(markup);
    /** @type {?} */
    let placeholderRegExp = /__([\w]+)__/g;
    /** @type {?} */
    let placeholders = {};
    /** @type {?} */
    let groups = /[\[\]()]+/.test(markup);
    /** @type {?} */
    let match;
    /** @type {?} */
    let i = groups ? 1 : 0;
    /** @type {?} */
    let placeholderExclusion = groups ? '^\\)\\]' : '\\S';
    do {
        match = placeholderRegExp.exec(markupPattern);
        if (match) {
            /** @type {?} */
            let placeholder = match[1];
            markupPattern = markupPattern.replace(`__${placeholder}__`, `([${placeholderExclusion}]+)`);
            placeholders[placeholder] = ++i;
        }
    } while (match);
    return { markup: markup, regEx: new RegExp('(' + markupPattern + ')', 'ig'), groups: placeholders };
}
/**
 * @param {?} value
 * @param {?} mentionMarkup
 * @param {?} displayTransform
 * @return {?}
 */
function getPlainText(value, mentionMarkup, displayTransform) {
    mentionMarkup.regEx.lastIndex = 0;
    return value.replace(mentionMarkup.regEx, displayTransform);
}
/**
 * @param {?} item
 * @param {?} markupMention
 * @return {?}
 */
function replacePlaceholders(item, markupMention) {
    /** @type {?} */
    let result = markupMention.markup + '';
    Object.keys(markupMention.groups)
        .forEach(key => result = result.replace(new RegExp(`__${key}__`, 'g'), typeof item === 'string' ? item : item[key]));
    return result;
}
/**
 * @param {?} value
 * @param {?} markupMention
 * @param {?} plainTextValue
 * @param {?=} selectionStartBeforeChange
 * @param {?=} selectionEndBeforeChange
 * @param {?=} selectionEndAfterChange
 * @param {?=} displayTransform
 * @return {?}
 */
function applyChangeToValue(value, markupMention, plainTextValue, selectionStartBeforeChange = 0, selectionEndBeforeChange = 0, selectionEndAfterChange = 0, displayTransform) {
    /** @type {?} */
    let oldPlainTextValue = getPlainText(value, markupMention, displayTransform);
    /** @type {?} */
    let lengthDelta = oldPlainTextValue.length - plainTextValue.length;
    if (!selectionEndBeforeChange) {
        selectionEndBeforeChange = selectionStartBeforeChange;
    }
    if (selectionStartBeforeChange === selectionEndBeforeChange && selectionEndBeforeChange === selectionEndAfterChange &&
        oldPlainTextValue.length === plainTextValue.length) {
        selectionStartBeforeChange--;
    }
    /** @type {?} */
    let insert = plainTextValue.slice(selectionStartBeforeChange, selectionEndAfterChange);
    /** @type {?} */
    let spliceStart = Math.min(selectionStartBeforeChange, selectionEndAfterChange);
    /** @type {?} */
    let spliceEnd = selectionEndBeforeChange;
    if (selectionStartBeforeChange === selectionEndAfterChange) {
        spliceEnd = Math.max(selectionEndBeforeChange, selectionStartBeforeChange + lengthDelta);
    }
    return spliceString(value, mapPlainTextIndex(value, markupMention, spliceStart, false, displayTransform), mapPlainTextIndex(value, markupMention, spliceEnd, true, displayTransform), insert);
}
/**
 * @param {?} value
 * @param {?} mentionMarkup
 * @param {?} indexInPlainText
 * @param {?} displayTransform
 * @return {?}
 */
function findStartOfMentionInPlainText(value, mentionMarkup, indexInPlainText, displayTransform) {
    /** @type {?} */
    let result = { start: -1, end: -1 };
    /** @type {?} */
    const markupIterator = (matchString, index, mentionPlainTextIndex, display) => {
        if (mentionPlainTextIndex < indexInPlainText && mentionPlainTextIndex + display.length > indexInPlainText) {
            result = { start: mentionPlainTextIndex, end: mentionPlainTextIndex + display.length };
            return true;
        }
        return false;
    };
    iterateOnlyMentionsMarkup(value, mentionMarkup, markupIterator, displayTransform);
    return result;
}
/**
 * @param {?} value
 * @param {?} mentionMarkup
 * @param {?} indexInPlainText
 * @param {?} displayTransform
 * @return {?}
 */
function getBoundsOfMentionAtPosition(value, mentionMarkup, indexInPlainText, displayTransform) {
    return findStartOfMentionInPlainText(value, mentionMarkup, indexInPlainText, displayTransform);
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class NgMentionsListComponent {
    constructor() {
        this.activeIndex = -1;
        this.itemSelected = new EventEmitter();
        this.show = false;
        this.dropUp = false;
        this._top = 0;
        this._left = 0;
    }
    /**
     * @return {?}
     */
    get selectedItem() {
        return this.activeIndex >= 0 && this.items[this.activeIndex] !== undefined ? this.items[this.activeIndex] : null;
    }
    /**
     * @return {?}
     */
    get top() {
        return this._top + this.adjustTop + 'px';
    }
    /**
     * @return {?}
     */
    get left() {
        return this._left + 'px';
    }
    /**
     * @return {?}
     */
    get noItems() {
        return !Array.isArray(this.items) || this.items.length === 0;
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        if (!this.itemTemplate) {
            this.itemTemplate = this.defaultItemTemplate;
        }
    }
    /**
     * @param {?} event
     * @param {?} activeIndex
     * @param {?} item
     * @return {?}
     */
    onItemClick(event, activeIndex, item) {
        event.preventDefault();
        this.activeIndex = activeIndex;
        this.itemSelected.emit(item);
    }
    /**
     * @return {?}
     */
    selectFirstItem() {
        this.activeIndex = 0;
        this.resetScroll();
    }
    /**
     * @return {?}
     */
    selectPreviousItem() {
        if (this.activeIndex > 0) {
            this.activeIndex--;
        }
        this.scrollToActiveItem();
    }
    /**
     * @return {?}
     */
    selectNextItem() {
        if (this.activeIndex < this.items.length - 1) {
            this.activeIndex++;
            this.scrollToActiveItem();
        }
    }
    /**
     * @return {?}
     */
    selectLastItem() {
        this.activeIndex = this.items.length > 0 ? this.items.length - 1 : 0;
        this.scrollToActiveItem();
    }
    /**
     * @return {?}
     */
    position() {
        /** @type {?} */
        const element = this.textAreaElement;
        /** @type {?} */
        let coords = getCaretCoordinates(element, element.selectionStart);
        console.log('coords.top', coords.top);
        console.log('this.textAreaElement.parentElement.offsetTop', this.textAreaElement.parentElement.offsetTop);
        this._top = coords.top + this.textAreaElement.parentElement.offsetTop;
        this._left = coords.left + element.offsetLeft;
        this.list.nativeElement.scrollTop = 0;
        setTimeout(() => {
            /** @type {?} */
            const rect = this.list.nativeElement.getBoundingClientRect();
            if (rect.x + rect.width > window.innerWidth) {
                /** @type {?} */
                const calcLeft = this._left - Math.abs(window.innerWidth - (rect.x + rect.width));
                this._left = calcLeft > 0 ? calcLeft : 0;
            }
        });
    }
    /**
     * @return {?}
     */
    resetScroll() {
        this.list.nativeElement.scrollTop = 0;
    }
    /**
     * @param {?} item
     * @return {?}
     */
    transformItem(item) {
        return this.displayTransform(item) || item;
    }
    /**
     * @return {?}
     */
    get adjustTop() {
        /** @type {?} */
        let adjust = 0;
        if (!this.dropUp) {
            /** @type {?} */
            const computed = getComputedStyle(this.textAreaElement);
            adjust = parseInt(computed.fontSize, 10) + this.textAreaElement.offsetTop;
        }
        return adjust;
    }
    /**
     * @return {?}
     */
    scrollToActiveItem() {
        /** @type {?} */
        let element = (/** @type {?} */ (this.list.nativeElement));
        setTimeout(() => {
            if (this.activeIndex === 0) {
                element.scrollTop = 0;
            }
            else {
                /** @type {?} */
                const activeElement = (/** @type {?} */ (element.querySelector('li.active')));
                if (activeElement) {
                    element.scrollTop = activeElement.offsetTop;
                }
            }
        });
    }
}
NgMentionsListComponent.decorators = [
    { type: Component, args: [{
                selector: 'mentions-list',
                template: `
      <ng-template #defaultItemTemplate let-item="item">
          {{transformItem(item)}}
      </ng-template>
      <ng-template #defaultHeaderTemplate let-item="item">
          {{item.type | titlecase}}s:
      </ng-template>
      <ul #list class="dropdown-menu scrollable-menu">
          <li *ngFor="let item of items; let i = index" [class.active]="activeIndex === i">
              <span class="dropdown-header" *ngIf="headerTemplate && (i === 0 || item.type !== items[i-1].type)">
                  <ng-template [ngTemplateOutlet]="headerTemplate || defaultHeaderTemplate"
                               [ngTemplateOutletContext]="{item:item,index:i}"></ng-template>
              </span>
              <a href class="dropdown-item" (mousedown)="onItemClick($event, i, item)" >
                  <ng-template [ngTemplateOutlet]="itemTemplate"
                               [ngTemplateOutletContext]="{item:item,index:i}"></ng-template>
              </a>
          </li>
      </ul>
  `,
                encapsulation: ViewEncapsulation.None,
                styles: ['mentions-list {position: absolute;display: none;}', 'mentions-list.drop-up {transform: translateY(-100%);}',
                    'mentions-list.show {display: block;} mentions-list.no-items {display: none;}',
                    'mentions-list .scrollable-menu {display: block;height: auto;max-height:300px;overflow:auto;}',
                    'mentions-list li.active a {background: #f7f7f9;}',
                    'mentions-list li .dropdown-header {display: block;}']
            }] }
];
NgMentionsListComponent.propDecorators = {
    defaultItemTemplate: [{ type: ViewChild, args: ['defaultItemTemplate',] }],
    defaultHeaderTemplate: [{ type: ViewChild, args: ['defaultHeaderTemplate',] }],
    list: [{ type: ViewChild, args: ['list',] }],
    show: [{ type: HostBinding, args: ['class.show',] }],
    dropUp: [{ type: HostBinding, args: ['class.drop-up',] }],
    top: [{ type: HostBinding, args: ['style.top',] }],
    left: [{ type: HostBinding, args: ['style.left',] }],
    noItems: [{ type: HostBinding, args: ['class.no-items',] }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/**
 * The Mentions Component
 */
class NgMentionsComponent {
    /**
     * @param {?} element
     * @param {?} componentResolver
     * @param {?} viewContainer
     * @param {?} changeDetector
     * @param {?} ngZone
     */
    constructor(element, componentResolver, viewContainer, changeDetector, ngZone) {
        this.element = element;
        this.componentResolver = componentResolver;
        this.viewContainer = viewContainer;
        this.changeDetector = changeDetector;
        this.ngZone = ngZone;
        /**
         * The character to trigger the mentions list when a user is typing in the input field
         */
        this.triggerChar = '@';
        this.includeTriggerChar = false;
        /**
         * The markup used to format a mention in the model value
         */
        this.mentionMarkup = '@[__display__](__type__:__id__)';
        /**
         * Optional. When using a custom search (i.e. an API call), the internal searching capability should be disabled.
         */
        this.disableSearch = false;
        /**
         * Only used when internal search is not disabled. This limits the maximum number of items to display in the search
         * result list.
         */
        this.maxItems = -1;
        /**
         * Used to cause the search result list to display in a "drop up" fashion, instead of a typical dropdown.
         */
        this.dropUp = false;
        /**
         * If the supplied mentions are a list of objects, this is the name of the property used to display
         * the mention in the search result list and when formatting a mention in the displayed text.
         */
        this.displayName = 'display';
        this.autoexpand = false;
        this.commitOnCMDEnter = false;
        /**
         * An event emitted, after the trigger character has been typed, with the user-entered search string.
         */
        this.search = new EventEmitter();
        this.valueChanges = new EventEmitter();
        this.stateChanges = new Subject();
        this.focusIn = new EventEmitter();
        this.focusLost = new EventEmitter();
        this.escapePressed = new EventEmitter();
        this.cmdEnter = new EventEmitter();
        this.focused = false;
        this.empty = false;
        this.displayContent = '';
        this.lines = [];
        this.highlighterStyle = {};
        this.textAreaClassNames = {};
        this.mentions = [];
        this.inputFallback = false;
        this._value = '';
        this._rows = 1;
        this._columns = 1;
        this.stopSearch = false;
        this.markupSearch = markupToRegExp(this.mentionMarkup);
        this._destroyed = new Subject();
        this.newLine = /\n/g;
        this._errorState = false;
    }
    /**
     * Classes for textarea
     * @return {?}
     */
    get formClass() {
        return Object.keys(this.textAreaClassNames).join(' ');
    }
    /**
     * @param {?} classNames
     * @return {?}
     */
    set formClass(classNames) {
        this.textAreaClassNames = {};
        Array.from(classNames.split(' ')).forEach(className => {
            this.textAreaClassNames[className] = true;
        });
    }
    /**
     * @return {?}
     */
    get value() {
        return this._value;
    }
    /**
     * @param {?} value
     * @return {?}
     */
    set value(value) {
        this.parseLines(value);
    }
    /**
     * @return {?}
     */
    get required() {
        return this._required;
    }
    /**
     * @param {?} value
     * @return {?}
     */
    set required(value) {
        this._required = value;
        this.refreshStyles();
    }
    /**
     * @return {?}
     */
    get disabled() {
        return this._disabled;
    }
    /**
     * @param {?} value
     * @return {?}
     */
    set disabled(value) {
        this._disabled = value;
        this.refreshStyles();
    }
    /**
     * @return {?}
     */
    get rows() {
        return this._rows;
    }
    /**
     * @param {?} value
     * @return {?}
     */
    set rows(value) {
        this._rows = Math.max(1, (/** @type {?} */ (value)));
        this.refreshStyles();
    }
    /**
     * @return {?}
     */
    get columns() {
        return this._columns;
    }
    /**
     * @param {?} value
     * @return {?}
     */
    set columns(value) {
        this._columns = Math.max(1, (/** @type {?} */ (value)));
        this.refreshStyles();
    }
    /**
     * The list of mentions to display, or filter, in the search result list.
     * @param {?} value
     * @return {?}
     */
    set mentionItems(value) {
        this.mentions = value;
        if (this.disableSearch && this.mentionsList) {
            this.mentionsList.items = value;
        }
    }
    /**
     * @return {?}
     */
    get readonly() {
        return this.disabled ? 'readonly' : null;
    }
    /**
     * @return {?}
     */
    get errorState() {
        return this._errorState;
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        this.parseMarkup();
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    ngOnChanges(changes) {
        if ('markup' in changes) {
            this.parseMarkup();
        }
    }
    /**
     * @return {?}
     */
    ngAfterViewInit() {
        this.parseLines(this._value);
        this.refreshStyles();
    }
    /**
     * @return {?}
     */
    ngAfterViewChecked() {
        this.refreshStyles();
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this._destroyed.next();
        this._destroyed.complete();
    }
    /**
     * @return {?}
     */
    onWindowResize() {
        this.refreshStyles();
    }
    /**
     * @return {?}
     */
    onTextAreaScroll() {
        this.highlighterElement.nativeElement.scrollTop = this.textAreaInputElement.nativeElement.scrollTop;
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onSelect(event) {
        this.selectionStart = event.target.selectionStart;
        this.selectionEnd = event.target.selectionEnd;
    }
    /**
     * @param {?} newPlainTextValue
     * @return {?}
     */
    onChange(newPlainTextValue) {
        /** @type {?} */
        let value = this._value;
        /** @type {?} */
        let displayTransform = this.displayTransform.bind(this);
        /** @type {?} */
        let selectionStart = this.textAreaInputElement.nativeElement.selectionStart;
        /** @type {?} */
        let selectionEnd = this.textAreaInputElement.nativeElement.selectionEnd;
        /** @type {?} */
        let bounds = getBoundsOfMentionAtPosition(newPlainTextValue, this.markupSearch, selectionStart, displayTransform);
        if (bounds.start !== -1) {
            newPlainTextValue = newPlainTextValue.substring(0, bounds.start) + newPlainTextValue.substring(bounds.end);
        }
        /** @type {?} */
        let newValue = applyChangeToValue(value, this.markupSearch, newPlainTextValue, this.selectionStart, this.selectionEnd, selectionEnd, displayTransform);
        /** @type {?} */
        let startOfMention = findStartOfMentionInPlainText(value, this.markupSearch, selectionStart, displayTransform);
        if (startOfMention.start > -1 && this.selectionEnd > startOfMention.start) {
            selectionStart = startOfMention.start;
            selectionEnd = selectionStart;
        }
        this.selectionStart = Math.max(selectionStart, 0);
        this.selectionEnd = Math.max(selectionEnd, 0);
        this.parseLines(newValue);
        if (this.selectionEnd > 0) {
            setTimeout(() => setCaretPosition(this.textAreaInputElement.nativeElement, this.selectionEnd));
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onInput(event) {
        if (this.inputFallback && event.data) {
            /** @type {?} */
            let characterPressed = event.data;
            /** @type {?} */
            let keyCode = characterPressed.charCodeAt(0);
            this.onKeyDown({ which: keyCode, keyCode: keyCode, key: characterPressed });
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onKeyDown(event) {
        /** @type {?} */
        let caretPosition = getCaretPosition(this.textAreaInputElement.nativeElement);
        if (this.inputFallback) {
            caretPosition--;
        }
        /** @type {?} */
        let characterPressed = event.key;
        /** @type {?} */
        let keyCode = event.which || event.keyCode;
        if (!characterPressed) {
            /** @type {?} */
            let characterCode = event.which || event.keyCode;
            characterPressed = String.fromCharCode(characterCode);
            if (!event.shiftKey && (characterCode >= 65 && characterCode <= 90)) {
                characterPressed = String.fromCharCode(characterCode + 32);
            }
        }
        this.inputFallback = keyCode === 229;
        if (this.inputFallback) {
            return;
        }
        if (event && event.stopPropagation) {
            event.stopPropagation();
        }
        if (keyCode === Key.Escape) {
            this.escapePressed.emit();
        }
        if (this.commitOnCMDEnter && keyCode === Key.Enter && event.metaKey) {
            this.cmdEnter.emit();
            event.preventDefault();
            return;
        }
        if (keyCode === Key.Enter && event.wasSelection && caretPosition < this.startPos) {
            caretPosition = this.startNode.length;
            setCaretPosition(this.startNode, caretPosition);
        }
        /** @type {?} */
        let startOfMention = findStartOfMentionInPlainText(this._value, this.markupSearch, caretPosition, this.displayTransform.bind(this));
        if (characterPressed === this.triggerChar) {
            this.setupMentionsList(caretPosition);
        }
        else if (startOfMention.start === -1 && this.startPos >= 0) {
            if (caretPosition <= this.startPos) {
                this.mentionsList.show = false;
                this.startPos = -1;
            }
            else if (keyCode !== Key.Shift && !event.shiftKey && !event.metaKey && !event.altKey && !event.ctrlKey &&
                caretPosition > this.startPos) {
                this.handleKeyDown(event, caretPosition, characterPressed);
            }
        }
        else {
            this.onSelect({ target: this.textAreaInputElement.nativeElement });
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onBlur(event) {
        if (event instanceof FocusEvent && event.relatedTarget) {
            /** @type {?} */
            let element = (/** @type {?} */ (event.relatedTarget));
            if (element.classList.contains('dropdown-item') || element.classList.contains('keep-focus')) {
                return;
            }
        }
        this.stopEvent(event);
        if (this.mentionsList) {
            this.mentionsList.show = false;
        }
        this.focused = false;
        this.focusLost.emit();
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onFocus(event) {
        this.focused = true;
        this.focusIn.emit(event);
    }
    /**
     * @param {?} part
     * @return {?}
     */
    isPartMention(part) {
        return typeof part.contents !== 'undefined';
    }
    /**
     * @param {?} mention
     * @return {?}
     */
    formatMention(mention) {
        return this._formatMention(mention.contents);
    }
    /**
     * @param {...?} _
     * @return {?}
     */
    displayTransform(..._) {
        /** @type {?} */
        let replaceIndex = this.markupSearch.groups[this.displayName];
        return (this.includeTriggerChar ? '@' : '') + _[replaceIndex];
    }
    /**
     * @param {?} contents
     * @return {?}
     */
    _formatMention(contents) {
        /** @type {?} */
        let replaceValue = `\$${this.displayName}`;
        /** @type {?} */
        let replaceIndex;
        /** @type {?} */
        let result = contents.replace(this.markupSearch.regEx, replaceValue);
        if (result === replaceValue) {
            replaceIndex = (this.includeTriggerChar ? '@' : '') + `\$${this.markupSearch.groups[this.displayName]}`;
            result = contents.replace(this.markupSearch.regEx, replaceIndex);
        }
        return result;
    }
    /**
     * @param {?} event
     * @return {?}
     */
    stopEvent(event) {
        if (event.preventDefault) {
            event.preventDefault();
            event.stopPropagation();
            event.stopImmediatePropagation();
        }
    }
    /**
     * @param {?} caretPosition
     * @return {?}
     */
    setupMentionsList(caretPosition) {
        this.startPos = caretPosition;
        this.startNode = window.getSelection().anchorNode;
        this.searchString = '';
        this.stopSearch = false;
        this.showMentionsList();
        this.updateMentionsList();
    }
    /**
     * @param {?} event
     * @param {?} caretPosition
     * @param {?} characterPressed
     * @return {?}
     */
    handleKeyDown(event, caretPosition, characterPressed) {
        /** @type {?} */
        let keyCode = event.which || event.keyCode;
        if (keyCode === Key.Space) {
            this.startPos = -1;
        }
        else if (keyCode === Key.Backspace && caretPosition > 0) {
            caretPosition--;
            if (caretPosition === this.startPos) {
                this.stopSearch = true;
            }
            this.mentionsList.show = !this.stopSearch;
        }
        else if (this.mentionsList.show) {
            if (keyCode === Key.Tab || keyCode === Key.Enter) {
                this.stopEvent(event);
                this.mentionsList.show = false;
                /** @type {?} */
                let value = this._value;
                /** @type {?} */
                let start = mapPlainTextIndex(value, this.markupSearch, this.startPos, false, this.displayTransform.bind(this));
                /** @type {?} */
                let item = event.item || this.mentionsList.selectedItem;
                /** @type {?} */
                let newValue = replacePlaceholders(item, this.markupSearch);
                /** @type {?} */
                let newDisplayValue = this._formatMention(newValue);
                caretPosition = this.startPos + newDisplayValue.length;
                /** @type {?} */
                let searchString = this.searchString || '';
                value = value.substring(0, start) + newValue + value.substring(start + searchString.length + 1, value.length);
                this.parseLines(value);
                this.startPos = -1;
                this.searchString = '';
                this.stopSearch = true;
                this.mentionsList.show = false;
                this.mentionsList.resetScroll();
                this.changeDetector.detectChanges();
                setTimeout(() => {
                    setCaretPosition(this.textAreaInputElement.nativeElement, caretPosition);
                    this.onSelect({ target: this.textAreaInputElement.nativeElement });
                });
                return;
            }
            else if (keyCode === Key.Escape) {
                this.stopEvent(event);
                this.mentionsList.show = false;
                this.stopSearch = true;
                return;
            }
            else if (keyCode === Key.ArrowDown) {
                this.stopEvent(event);
                this.mentionsList.selectNextItem();
                return;
            }
            else if (keyCode === Key.ArrowUp) {
                this.stopEvent(event);
                this.mentionsList.selectPreviousItem();
                return;
            }
            else if (keyCode === Key.Home) {
                this.stopEvent(event);
                this.mentionsList.selectFirstItem();
                return;
            }
            else if (keyCode === Key.End) {
                this.stopEvent(event);
                this.mentionsList.selectLastItem();
                return;
            }
        }
        if (keyCode === Key.ArrowLeft || keyCode === Key.ArrowRight || keyCode === Key.Home || keyCode === Key.End) {
            this.onSelect(event);
            return;
        }
        /** @type {?} */
        let mention = this.displayContent.substring(this.startPos + 1, caretPosition);
        if (keyCode !== Key.Backspace) {
            mention += characterPressed;
        }
        this.searchString = mention || '';
        this.updateMentionsList();
    }
    /**
     * @param {?} item
     * @return {?}
     */
    getDisplayValue(item) {
        if (typeof item === 'string') {
            return item;
        }
        else if (item[this.displayName] !== undefined) {
            return item[this.displayName];
        }
        return null;
    }
    /**
     * @return {?}
     */
    showMentionsList() {
        if (!this.mentionsList) {
            /** @type {?} */
            let componentFactory = this.componentResolver.resolveComponentFactory(NgMentionsListComponent);
            /** @type {?} */
            let componentRef = this.viewContainer.createComponent(componentFactory);
            this.mentionsList = componentRef.instance;
            this.mentionsList.itemTemplate = this.mentionListItemTemplate;
            this.mentionsList.headerTemplate = this.mentionListHeaderTemplate;
            this.mentionsList.displayTransform = this.displayTransform.bind(this);
            this.mentionsList.itemSelected.subscribe(item => {
                this.textAreaInputElement.nativeElement.focus();
                /** @type {?} */
                const fakeEvent = { keyCode: Key.Enter, wasSelection: true, item: item };
                this.onKeyDown(fakeEvent);
            });
            this.mentionsList.displayTransform = this.getDisplayValue.bind(this);
        }
        this.mentionsList.textAreaElement = this.textAreaInputElement.nativeElement;
        this.mentionsList.show = true;
        this.mentionsList.dropUp = this.dropUp;
        this.mentionsList.activeIndex = 0;
        this.mentionsList.position();
        this.ngZone.run(() => this.mentionsList.resetScroll());
    }
    /**
     * @return {?}
     */
    updateMentionsList() {
        if (!this.disableSearch) {
            /** @type {?} */
            let items = Array.from(this.mentions);
            if (this.searchString) {
                /** @type {?} */
                let searchString = this.searchString.toLowerCase();
                /** @type {?} */
                let searchRegEx = new RegExp(escapeRegExp(searchString), 'i');
                items = items.filter(item => {
                    /** @type {?} */
                    let value = this.getDisplayValue(item);
                    return value !== null && searchRegEx.test(value);
                });
                if (this.maxItems > 0) {
                    items = items.slice(0, this.maxItems);
                }
            }
            this.mentionsList.items = items;
        }
        else {
            this.search.emit(this.searchString);
        }
    }
    /**
     * @return {?}
     */
    parseMarkup() {
        if (this.mentionMarkup.length === 0 || this.mentionMarkup[0] !== this.triggerChar) {
            throw new Error(`ng-mentions markup pattern must start with the trigger character "${this.triggerChar}"`);
        }
        this.markupSearch = markupToRegExp(this.mentionMarkup);
    }
    /**
     * @param {?=} value
     * @return {?}
     */
    parseLines(value = '') {
        if (value !== this._value) {
            value = value || '';
            /** @type {?} */
            let lines = value.split(this.newLine).map((line) => this.formatMentions(line));
            /** @type {?} */
            let displayContent = lines.map(line => line.content).join('\n');
            if (this.displayContent !== displayContent) {
                this.lines = lines;
                this.displayContent = displayContent;
                this.triggerChange(value);
            }
        }
        this.empty = !value;
    }
    /**
     * @param {?} line
     * @return {?}
     */
    formatMentions(line) {
        /** @type {?} */
        let lineObj = (/** @type {?} */ ({ originalContent: line, content: line, parts: [] }));
        if (line.length === 0) {
            return lineObj;
        }
        /** @type {?} */
        let tags = [];
        /** @type {?} */
        let mention;
        /** @type {?} */
        let regEx = this.markupSearch.regEx;
        regEx.lastIndex = 0;
        while ((mention = regEx.exec(line)) !== null) {
            tags.push({ indices: { start: mention.index, end: mention.index + mention[0].length } });
        }
        /** @type {?} */
        let prevTags = [];
        /** @type {?} */
        let content = '';
        [...tags].sort((tagA, tagB) => tagA.indices.start - tagB.indices.start).forEach((tag) => {
            /** @type {?} */
            const expectedLength = tag.indices.end - tag.indices.start;
            /** @type {?} */
            const contents = line.slice(tag.indices.start, tag.indices.end);
            if (contents.length === expectedLength) {
                /** @type {?} */
                const prevIndex = prevTags.length > 0 ? prevTags[prevTags.length - 1].indices.end : 0;
                /** @type {?} */
                const before = line.slice(prevIndex, tag.indices.start);
                /** @type {?} */
                const partMention = (/** @type {?} */ ({ contents: contents, tag: tag }));
                lineObj.parts.push(before);
                lineObj.parts.push(partMention);
                prevTags.push(tag);
                content += before + this.formatMention(partMention);
            }
        });
        /** @type {?} */
        const remainingStart = prevTags.length > 0 ? prevTags[prevTags.length - 1].indices.end : 0;
        /** @type {?} */
        const remaining = line.slice(remainingStart);
        lineObj.parts.push(remaining);
        content += remaining;
        lineObj.content = content;
        return lineObj;
    }
    /**
     * @return {?}
     */
    refreshStyles() {
        /** @type {?} */
        let element = this.textAreaInputElement.nativeElement;
        if (this.autoexpand) {
            // Reset textarea height to auto that correctly calculate the new height
            element.style.height = 'auto';
            // Set new height
            element.style.height = `${element.scrollHeight}px`;
        }
        /** @type {?} */
        let computedStyle = getComputedStyle(element);
        this.highlighterStyle = {};
        styleProperties.forEach(prop => {
            this.highlighterStyle[prop] = computedStyle[prop];
        });
        this.changeDetector.detectChanges();
    }
    /**
     * @param {?} value
     * @return {?}
     */
    triggerChange(value) {
        this._value = value;
        this.valueChanges.emit(this._value);
        this.changeDetector.detectChanges();
    }
}
NgMentionsComponent.decorators = [
    { type: Component, args: [{
                exportAs: 'ngMentions',
                selector: 'ng-mentions',
                template: "<div #highlighter class=\"highlighter\" [ngClass]=\"textAreaClassNames\" [attr.readonly]=\"readonly\"\n     [ngStyle]=\"highlighterStyle\">\n  <div *ngFor=\"let line of lines\" class=\"lines\" [class.has-parsed-content]=\"parsedContent\">\n    <ng-container *ngFor=\"let part of line.parts\">\n      <highlighted *ngIf=\"isPartMention(part)\" [tag]=\"part.tag\">{{formatMention(part)}}</highlighted>\n      <ng-container *ngIf=\"!isPartMention(part)\">{{part}}</ng-container>\n    </ng-container>\n    &nbsp;\n  </div>\n  <div class=\"parsed-content\" (click)=\"input.focus()\" [hidden]=\"!parsedContent\" [innerHTML]=\"parsedContent\"></div>\n</div>\n<textarea\n  #input\n  [rows]=\"rows\"\n  [cols]=\"columns\"\n  [ngModel]=\"displayContent\"\n  [ngClass]=\"textAreaClassNames\"\n  (keydown)=\"onKeyDown($event)\"\n  (input)=\"onInput($event)\"\n  (blur)=\"onBlur($event)\"\n  (select)=\"onSelect($event)\"\n  (focus)=\"onFocus($event)\"\n  (mouseup)=\"onSelect($event)\"\n  (ngModelChange)=\"onChange($event)\"\n  (scroll)=\"onTextAreaScroll()\"\n  [disabled]=\"disabled\"\n  [required]=\"required\"\n  [placeholder]=\"placeholder\"\n></textarea>\n",
                preserveWhitespaces: false,
                encapsulation: ViewEncapsulation.None,
                styles: ["ng-mentions{position:relative}ng-mentions textarea{position:relative;background-color:transparent!important;opacity:0}ng-mentions .highlighter{position:absolute;top:0;left:0;right:0;bottom:0;overflow:hidden!important}ng-mentions .highlighter .lines{white-space:pre-wrap}ng-mentions .highlighter .lines.has-parsed-content{display:none}ng-mentions .highlighter .parsed-content{display:block}ng-mentions highlighted{display:inline;border-radius:3px;padding:1px;margin:-1px;overflow-wrap:break-word}ng-mentions.focused .highlighter highlighted{background-color:#add8e6;color:transparent}ng-mentions.focused .highlighter .lines,ng-mentions.focused .highlighter .lines.has-parsed-content{color:transparent;display:block}ng-mentions.focused .highlighter .parsed-content{display:none}ng-mentions.empty textarea,ng-mentions.focused textarea{opacity:1}"]
            }] }
];
/** @nocollapse */
NgMentionsComponent.ctorParameters = () => [
    { type: ElementRef },
    { type: ComponentFactoryResolver },
    { type: ViewContainerRef },
    { type: ChangeDetectorRef },
    { type: NgZone }
];
NgMentionsComponent.propDecorators = {
    triggerChar: [{ type: Input, args: ['triggerChar',] }],
    includeTriggerChar: [{ type: Input, args: ['includeTriggerChar',] }],
    mentionMarkup: [{ type: Input, args: ['markup',] }],
    disableSearch: [{ type: Input, args: ['disableSearch',] }],
    maxItems: [{ type: Input, args: ['maxItems',] }],
    dropUp: [{ type: Input, args: ['dropUp',] }],
    displayName: [{ type: Input, args: ['displayName',] }],
    formClass: [{ type: Input, args: ['formClass',] }],
    placeholder: [{ type: Input, args: ['placeholder',] }],
    value: [{ type: Input, args: ['value',] }],
    required: [{ type: Input, args: ['required',] }],
    disabled: [{ type: Input, args: ['disabled',] }],
    rows: [{ type: Input, args: ['rows',] }],
    columns: [{ type: Input, args: ['cols',] }],
    mentionItems: [{ type: Input, args: ['mentions',] }],
    parsedContent: [{ type: Input }],
    autoexpand: [{ type: Input }],
    commitOnCMDEnter: [{ type: Input }],
    mentionListItemTemplate: [{ type: Input }],
    mentionListHeaderTemplate: [{ type: Input }],
    search: [{ type: Output, args: ['search',] }],
    valueChanges: [{ type: Output, args: ['valueChanges',] }],
    stateChanges: [{ type: Output, args: ['stateChanges',] }],
    focusIn: [{ type: Output, args: ['focusIn',] }],
    focusLost: [{ type: Output, args: ['focusLost',] }],
    escapePressed: [{ type: Output, args: ['escapePressed',] }],
    cmdEnter: [{ type: Output, args: ['cmdEnter',] }],
    textAreaInputElement: [{ type: ViewChild, args: ['input',] }],
    highlighterElement: [{ type: ViewChild, args: ['highlighter',] }],
    focused: [{ type: HostBinding, args: ['class.focused',] }],
    empty: [{ type: HostBinding, args: ['class.empty',] }],
    onWindowResize: [{ type: HostListener, args: ['window:resize',] }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/**
 * The NgMentionsAccessorDirective directive is used to indicate the input element.
 */
class NgMentionsAccessorDirective {
    /**
     * @param {?} element
     * @param {?} host
     */
    constructor(element, host) {
        this.element = element;
        this.host = host;
        this._destroyed = new Subject();
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        this.host.valueChanges.pipe(takeUntil(this._destroyed)).subscribe(value => this.onChange(value));
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this._destroyed.next();
        this._destroyed.complete();
    }
    /**
     * @param {?} fn
     * @return {?}
     */
    registerOnChange(fn) {
        this._onChange = fn;
    }
    /**
     * @param {?} fn
     * @return {?}
     */
    registerOnTouched(fn) {
        this._onTouch = fn;
    }
    /**
     * @param {?} isDisabled
     * @return {?}
     */
    setDisabledState(isDisabled) {
        this.host.disabled = isDisabled;
    }
    /**
     * @param {?} value
     * @return {?}
     */
    writeValue(value) {
        if (typeof value === 'string' || value === null) {
            this.host.value = value;
        }
    }
    /**
     * @param {?} value
     * @return {?}
     */
    onChange(value) {
        if (this._onChange) {
            this._onChange(value);
        }
    }
    /**
     * @return {?}
     */
    onTouched() {
        if (this._onTouch) {
            this._onTouch();
        }
    }
}
NgMentionsAccessorDirective.decorators = [
    { type: Directive, args: [{
                exportAs: 'ngMentions',
                selector: 'ng-mentions',
                host: { '(change)': 'onChange($event)', '(touch)': 'onTouched()' },
                providers: [{ provide: NG_VALUE_ACCESSOR, useExisting: forwardRef(() => NgMentionsAccessorDirective), multi: true }]
            },] }
];
/** @nocollapse */
NgMentionsAccessorDirective.ctorParameters = () => [
    { type: ElementRef },
    { type: NgMentionsComponent }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class HighlightedDirective {
}
HighlightedDirective.decorators = [
    { type: Directive, args: [{ selector: 'highlighted' },] }
];
HighlightedDirective.propDecorators = {
    tag: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/** @type {?} */
const EXPORT_DIRECTIVES = [NgMentionsComponent, NgMentionsAccessorDirective];
/** @type {?} */
const DECLARATIONS = [NgMentionsComponent, NgMentionsAccessorDirective, NgMentionsListComponent, HighlightedDirective];
class NgMentionsModule {
}
NgMentionsModule.decorators = [
    { type: NgModule, args: [{
                imports: [CommonModule, FormsModule],
                exports: EXPORT_DIRECTIVES,
                declarations: DECLARATIONS,
                entryComponents: [NgMentionsListComponent]
            },] }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */

export { NgMentionsAccessorDirective, NgMentionsComponent, NgMentionsModule, HighlightedDirective as ɵb, NgMentionsListComponent as ɵa };

//# sourceMappingURL=ng-mentions.js.map