import { AfterViewChecked, AfterViewInit, ChangeDetectorRef, ComponentFactoryResolver, ElementRef, EventEmitter, NgZone, OnChanges, OnDestroy, OnInit, SimpleChanges, TemplateRef, ViewContainerRef } from '@angular/core';
import { Subject } from 'rxjs';
import { Line, Mention } from './util/interfaces';
import { NgMentionsListComponent } from './util/mentions-list.component';
import { MarkupMention } from './util/utils';
/**
 * The Mentions Component
 */
export declare class NgMentionsComponent implements OnChanges, OnInit, AfterViewInit, AfterViewChecked, OnDestroy {
    private element;
    private componentResolver;
    private viewContainer;
    private changeDetector;
    private ngZone;
    /**
     * The character to trigger the mentions list when a user is typing in the input field
     */
    triggerChar: string;
    includeTriggerChar: boolean;
    /**
     * The markup used to format a mention in the model value
     */
    mentionMarkup: string;
    /**
     * Optional. When using a custom search (i.e. an API call), the internal searching capability should be disabled.
     */
    disableSearch: boolean;
    /**
     * Only used when internal search is not disabled. This limits the maximum number of items to display in the search
     * result list.
     */
    maxItems: number;
    /**
     * Used to cause the search result list to display in a "drop up" fashion, instead of a typical dropdown.
     */
    dropUp: boolean;
    /**
     * If the supplied mentions are a list of objects, this is the name of the property used to display
     * the mention in the search result list and when formatting a mention in the displayed text.
     */
    displayName: string;
    /**
     * Classes for textarea
     */
    formClass: string;
    placeholder: string;
    value: string;
    required: boolean;
    disabled: boolean;
    rows: number | string;
    columns: number | string;
    /**
     * The list of mentions to display, or filter, in the search result list.
     */
    mentionItems: any[];
    parsedContent: string;
    autoexpand: boolean;
    commitOnCMDEnter: boolean;
    mentionListItemTemplate: TemplateRef<any>;
    mentionListHeaderTemplate: TemplateRef<any>;
    /**
     * An event emitted, after the trigger character has been typed, with the user-entered search string.
     */
    readonly search: EventEmitter<string>;
    readonly valueChanges: EventEmitter<string>;
    readonly stateChanges: Subject<void>;
    readonly focusIn: EventEmitter<void>;
    readonly focusLost: EventEmitter<void>;
    readonly escapePressed: EventEmitter<void>;
    readonly cmdEnter: EventEmitter<void>;
    textAreaInputElement: ElementRef;
    highlighterElement: ElementRef;
    focused: boolean;
    empty: boolean;
    displayContent: string;
    lines: Line[];
    highlighterStyle: {
        [key: string]: string;
    };
    textAreaClassNames: {
        [key: string]: boolean;
    };
    selectionStart: number;
    selectionEnd: number;
    mentions: any[];
    inputFallback: boolean;
    readonly readonly: string;
    readonly errorState: boolean;
    private _value;
    private _required;
    private _disabled;
    private _rows;
    private _columns;
    private searchString;
    private startPos;
    private startNode;
    mentionsList: NgMentionsListComponent;
    private stopSearch;
    markupSearch: MarkupMention;
    private _destroyed;
    private newLine;
    private _errorState;
    constructor(element: ElementRef, componentResolver: ComponentFactoryResolver, viewContainer: ViewContainerRef, changeDetector: ChangeDetectorRef, ngZone: NgZone);
    ngOnInit(): void;
    ngOnChanges(changes: SimpleChanges): void;
    ngAfterViewInit(): void;
    ngAfterViewChecked(): void;
    ngOnDestroy(): void;
    onWindowResize(): void;
    onTextAreaScroll(): void;
    onSelect(event: any): void;
    onChange(newPlainTextValue: string): void;
    onInput(event: any): void;
    onKeyDown(event: any): void;
    onBlur(event: MouseEvent | KeyboardEvent): void;
    onFocus(event: any): void;
    isPartMention(part: any): boolean;
    formatMention(mention: Mention): string;
    private displayTransform;
    private _formatMention;
    private stopEvent;
    private setupMentionsList;
    private handleKeyDown;
    private getDisplayValue;
    private showMentionsList;
    private updateMentionsList;
    private parseMarkup;
    private parseLines;
    private formatMentions;
    private refreshStyles;
    private triggerChange;
}
